; ---purpose---
; Convert arithmetic expressions in infix notation
; to S-expressions. Infix expressions are represented
; by flat lists of variables (atoms) operators (atoms)
; and numbers (sequences of atoms representing digits).
; --
; The following operators are recognized: + (addition)
; - (subtraction), * (multiplication), / (division),
; ^ (exponentation), and - (negation). Brackets are
; recoginzed as parentheses.
; --
; For example,
; -
; (PREFIX '#57*[X+Y])
; -
; gives
; -
; '(ITIMES (QUOTE #+57) (IPLUS X Y)).

; ---keywords---
; PREFIX function, infix to prefix, RPN, conversion

; ---args---
; X - arithmetic expression

; ---example---
; (PREFIX '#12+34*56^[7+8])

:L LIB/DIGITS.L
:L LIB/REVERSE.L
:L LIB/MEMBER.L
:L LIB/CAAR.L
;L LIB/CADR.L
:L LIB/CAAAR.L
;L LIB/CAADR.L
;L LIB/CDADR.L
:L LIB/LOGNOT.L
:L LIB/IPLUS.L
:L LIB/IDIFFERENCE.L
:L LIB/ITIMES.L
:L LIB/IQUOTIENT.L
:L LIB/IEXPT.L

; ---code---
(DEFUN PREFIX (X)
	(LABEL

	; Operator symbols
	((+ '+) (- '-) (* '*) (/ '/) (^ '^) ([ '[) (] '])

	(SYMBOLP (LAMBDA (X)
		(MEMBER X '#ABCDEFGHIJKLMNOPQRSTUVWXYZ)))

	(NUMERICP (LAMBDA (X)
		(COND ((ATOM X) *F*) (T (EQ (CAR X) 'QUOTE)))))

	(OPERATORP (LAMBDA (X) (MEMBER X '#^*/+-)))

	(DIGITP (LAMBDA (X) (MEMBER X DIGITS)))

	; Extract a numeric value from the beginning of X.
	; Return (N XR) where N is the value extracted and
	; XR is the rest of X (X with N removed).
	(NUMBER (LAMBDA (X R)
		(COND ((NULL X)
				(LIST (LIST QUOTE (CONS + (REVERSE R))) X))
			((LOGNOT (DIGITP (CAR X)))
				(LIST (LIST QUOTE (CONS + (REVERSE R))) X))
			(T (NUMBER (CDR X) (CONS (CAR X) R))))))

	; These functions are used to extract parts
	; of (EXPR REST) lists where EXPR is the prefix
	; expression built so far and REST is the rest
	; the source expression to parse
	(CAR-OF-REST CAADR)
	(CDR-OF-REST CDADR)
	(EXPR CAR)
	(REST CADR)

	; Parse factors:
	; factor := [ sum ]
	; 	| - factor
	;	| Number
	;	| Symbol
	(FACTOR (LAMBDA (X)
		(COND ((NULL X) (LIST X NIL))
			; Parse parenthesized subexpressions
			((EQ (CAR X) [)
				(LABEL ((XSUB (SUM (CDR X))))
					(COND ((NULL (REST XSUB)) (BOTTOM 'MISSING-PAREN))
						((EQ (CAR-OF-REST XSUB) ])
							(LIST (EXPR XSUB) (CDR-OF-REST XSUB)))
						(T (BOTTOM 'MISSING-PAREN)))))
			; Parse applications of unary minuses
			((EQ (CAR X) -)
				(LABEL ((FAC (FACTOR (CDR X))))
					(LIST (LIST 'MINUS (EXPR FAC)) (REST FAC))))
			; Parse literal numbers
			((DIGITP (CAR X))
				(NUMBER X NIL))
			; Parse symbols
			(T (LIST (CAR X) (CDR X))))))

	; Parse powers:
	; power := factor
	;	| factor ^ power
	(POWER (LAMBDA (X)
		(LABEL ((LEFT (FACTOR X)))
		(COND ((NULL (REST LEFT)) LEFT)
			((EQ (CAR-OF-REST LEFT) ^)
				(LABEL ((RIGHT (POWER (CDR-OF-REST LEFT))))
					(LIST (LIST 'IEXPT (EXPR LEFT) (EXPR RIGHT))
						(REST RIGHT))))
			(T LEFT)))))

	; Parse terms:
	; term := power
	;	| power Symbol
	;	| power * term
	;	| power / term
	(TERM (LAMBDA (X)
		(LABEL ((LEFT (POWER X)))
		(COND ((NULL (REST LEFT)) LEFT)
			((SYMBOLP (CAR-OF-REST LEFT))
				(LABEL ((RIGHT (TERM (REST LEFT))))
					(LIST (LIST 'ITIMES (EXPR LEFT) (EXPR RIGHT))
						(REST RIGHT))))
			((EQ (CAR-OF-REST LEFT) *)
				(LABEL ((RIGHT (TERM (CDR-OF-REST LEFT))))
					(LIST (LIST 'ITIMES (EXPR LEFT) (EXPR RIGHT))
						(REST RIGHT))))
			((EQ (CAR-OF-REST LEFT) /)
				(LABEL ((RIGHT (TERM (CDR-OF-REST LEFT))))
					(LIST (LIST 'IQUOTIENT (EXPR LEFT) (EXPR RIGHT))
						(REST RIGHT))))
			(T LEFT)))))

	; Parse sums:
	; sum := term
	;	| term + sum
	;	| term - sum
	(SUM (LAMBDA (X)
		(LABEL ((LEFT (TERM X)))
		(COND ((NULL (REST LEFT)) LEFT)
			((EQ (CAR-OF-REST LEFT) +)
				(LABEL ((RIGHT (SUM (CDR-OF-REST LEFT))))
					(LIST (LIST 'IPLUS (EXPR LEFT) (EXPR RIGHT))
						(REST RIGHT))))
			((EQ (CAR-OF-REST LEFT) -)
				(LABEL ((RIGHT (SUM (CDR-OF-REST LEFT))))
					(LIST (LIST 'IDIFFERENCE (EXPR LEFT) (EXPR RIGHT))
						(REST RIGHT))))
			(T LEFT))))))

	; Pass X to the recursive descent parser consisting of
	; SUM, TERM, POWER, FACTOR. The parsing process returns a
	; list of the form (EXPR REST) as described above. When the
	; REST is NIL, the entire expression could be parsed
	; successfully.
	(LABEL ((PX (SUM X)))
		(COND ((LOGNOT (NULL (REST PX)))
				(BOTTOM (LIST 'SYNTAX: (CADR PX))))
			(T (EXPR PX))))))

