; ---purpose---
; A simple PROLOG interpreter.
; --
; This program is heavily based on a tiny PROLOG
; interpreter in MACLISP written by Ken Kahn,
; which was inspired by other tiny LISP-based
; PROLOGs of Par Emanuelson and Martin Nilsson.
; --
; This interpreter is completely free of side
; effects. It does not even perform any I/O, but
; returns its query results in a list.

; ---args---
; Q - query
; DB - database

; ---keywords---
; PROLOG interpreter, tiny PROLOG, database, unification, logic

; ---example---
; ; Sample database:
; (DEFINE DATABASE '(
;		((MALE ANTHONY))			; Gender facts
;		((MALE BERTRAM))
;		((FEMALE CATHY))
;		((FEMALE DENISE))
;		((MALE ERIC))
;		((FEMALE FANNY))
;
;		((PARENT BERTRAM ERIC))		; Parent relationship facts
;		((PARENT CATHY ERIC))
;		((PARENT ANTHONY CATHY))
;		((PARENT ERIC DENISE))
;		((PARENT ANTHONY FANNY))
;
;		((MOTHER (? A) (? B))		; A is mother of B if
;			(FEMALE (? A))			;  A is female and
;			(PARENT (? A) (? B)))	;  A is parent of B
;
;		((FATHER (? A) (? B))		; A is father of B if
;			(MALE (? A))			;  A is male and
;			(PARENT (? A) (? B)))	;  A is parent of B
;
;		((WIFE (? A) (? B))			; A is (in many cases) wife of B if
;			(MOTHER (? A) (? X))	;  A is mother of X and
;			(FATHER (? B) (? X)))	;  B is father of X
;
;		((CHILD (? A) (? B))		; A is child of B if
;			(PARENT (? B) (? A)))	;  B is parent of A
;
;		((DESCENDANT (? A) (? B))	; A is descendant of B if
;			(CHILD (? A) (? B)))	;  A is child of B
;
;		((DESCENDANT (? A) (? B))	; or if
;			(CHILD (? A) (? X))		;  A is child of X and
;			(DESCENDANT (? X) (? B)))	;  X is descendant of B
; ))
;
; ; Don't forget to enable trace output
; :T PRINT
;
; ; Sample queries:
; (QUERY '(FATHER ANTHONY (? CHILD)))			; whose father is Anthony?
; (QUERY '(PARENT (? PARENT) ERIC))			; who are Eric's parents?
; (QUERY '(DESCENDANT (? DESCENDANT) BERTRAM))	; list descendants of Bertram.
; (QUERY '(WIFE (? WIFE) (? HUSBAND)))		; who is who's wife?
; (QUERY '((? RELATION) CATHY (? PERSON)))	; which relations does Cathy have?

:L LIB/APPEND.L
:L LIB/ASSOC.L
:L LIB/EQUAL.L
:L LIB/CAAR.L
;L LIB/CADR.L
:L LIB/CAAAR.L
;L LIB/CADDR.L

; ---code---
(DEFUN PROLOG (Q DB)
	(LABEL

	; Is X a variable?
	; Variables are lists beginning with '?
	((VARP (LAMBDA (X)
		(COND ((ATOM X) *F*)
			(T (EQ (CAR X) '?)))))

	; Create an environment of unique names by
	; appending ID to each variable.
	(NEWSCOPE (LAMBDA (X ID)
		(LABEL ((NSC (LAMBDA (X)
			(COND ((ATOM X) X)
				((VARP X) (APPEND X ID))
				(T (CONS (NSC (CAR X)) (NSC (CDR X))))))))
		(NSC X))))

	; Increment ID.
	(NEXTID (LAMBDA (X) (LIST (CONS 'I (CAR X)))))

	; Find the value of a variable in a given environment.
	(VALUE (LAMBDA (X ENV)
		(COND ((VARP X)
				(LABEL ((B (ASSOC X ENV)))
					; If the variable is not bound, return its name
					(COND ((NULL B) X)
						; else return the value of its value
						; (variables may be bound to variables).
						(T (VALUE (CDR B) ENV)))))
			; Pass thru non-variables.
			(T X))))

	; Unify two expressions X and Y in a given environment.
	; If X and Y can be unified, return a new environment in
	; which the variables of X and Y are bound.
	; If the unification fails, return NIL.
	(UNIFY (LAMBDA (X Y ENV)
		; First retrieve the values of the expressions.
		(LABEL ((X (VALUE X ENV))
			(Y (VALUE Y ENV)))
		(COND
			; Variables are unified by binding them to
			; the other expression.
			((VARP X) (CONS (CONS X Y) ENV))
			((VARP Y) (CONS (CONS Y X) ENV))
			; Only identical atoms can be unified.
			((ATOM X) (COND ((EQ X Y) ENV) (T NIL)))
			((ATOM Y) (COND ((EQ X Y) ENV) (T NIL)))
			; Rules are unified by unifying their heads,
			; giving a new environment.
			; If the new environment is non-NIL, the rest
			; of the rules is unified in the new environment.
			(T (LABEL ((NEWENV (UNIFY (CAR X) (CAR Y) ENV)))
				(COND ((NULL NEWENV) NIL)	; Heads could not be unified.
					(T (UNIFY (CDR X) (CDR Y) NEWENV)))))))))

	; Attempt to unify each goal with each rule.
	; RULES  = list of rules (database)
	; GOALS  = conjunction of goals
	; ENV    = environment to unify in
	; ID     = scope ID
	; RESULT = list of results so far
	(TRYRULES (LAMBDA (RULES GOALS ENV ID RESULT)
		(COND ((NULL RULES) RESULT)	; Out of rules: return result.
			(T (LABEL
				; Create new scope for the current rule
				((THISRULE (NEWSCOPE (CAR RULES) ID))
				; and attempt to unify the current goal with the
				; head of this rule.
				(NEWENV (UNIFY (CAR GOALS) (CAR THISRULE) ENV)))
				(COND ((NULL NEWENV)
						; Unification failed, try next rule.
						(TRYRULES (CDR RULES) GOALS ENV ID RESULT))
					; Unification succeeded, prove union of rest of
					; this rule and remainder of goals in NEWENV.
					(T (LABEL ((RES (PROVE (APPEND (CDR THISRULE) (CDR GOALS))
							NEWENV (NEXTID ID))))
						; Add the result of above proof to the
						; result list and advance to the next rule.
						(TRYRULES (CDR RULES) GOALS ENV ID
							(APPEND RESULT RES))))))))))

	; Pretty-print results using :T meta-command
	(PRINT (LAMBDA (X) X))

	; Create an N-tuple of bindings
	; ( (VAR-1 = VALUE-1) ... (VAR-N = VALUE-N) ).
	; Each binding is prepresented by
	; (VAR = VALUE),
	; where VAR is the name and VALUE is the value
	; of a variable bound in the outermost scope
	; (the scope of the query).
	; N is the number of variables in the query.
	(LISTENV (LAMBDA (ENV)
		(LABEL ((LSE (LAMBDA (E RES)
			(COND ((NULL (CDR E)) (LIST (PRINT RES)))
				; Variables of the outer scope have ID=NIL
				((NULL (CADDR (CAAR E)))	; ID=NIL?
					; Move to next variable
					(LSE (CDR E)
						; adding a new binding to the result list
						(CONS (LIST (CADR (CAAR E))
								'= (VALUE (CAAR E) ENV))
							RES)))
				; Inner scope, skip
				(T (LSE (CDR E) RES))))))
		(LSE ENV NIL))))

	; Prove a list of goals in a given environment.
	; GOALS = list of goals
	; ENV   = environment
	; ID    = scope ID, see NEWSCOPE above
	(PROVE (LAMBDA (GOALS ENV ID)
		(COND ((NULL GOALS) (LISTENV ENV))	
			(T (TRYRULES DB GOALS ENV ID NIL))))))

	; ((NIL)) is the initial environment. It is used
	; because NIL would indicate failure in UNIFY.
	(PROVE (LIST (NEWSCOPE Q '(()))) '((NIL)) '((I)))))

; This is a utility function to submit queries to
; a predefined database bound to DATABASE.
; The result is discarded.
; QUERY requires that the PRINT function be traced
; (use the :T meta command to enable tracing).
(DEFUN QUERY (Q)
	(LABEL ((RESULT (PROLOG Q DATABASE)))
		(COND ((EQUAL RESULT '(())) 'YES)
			(T 'NO))))

