; ---purpose---
; Evaluate a pure LISP expression in a given environment.
; This function is rather a model than an implementation.
; --
; Primitives supported:
; ATOM BOTTOM CAR CDR COND CONS EQ EXPLODE IMPLODE LABEL
; LAMBDA LIST NULL QUOTE.
; --
; Internally defined symbols: *F* *T* NIL T.

; ---keywords---
; XEVAL function, EVAL, LISP semantics, formal semantics
; evaluate, evaluation

; ---args---
; X - s-expression to evaluate
; E - list of (name.value) pairs.

; ---example---
; ; Evaluate the Towers of Hanoi
; (XEVAL '(HANOI (QUOTE A) (QUOTE B) (QUOTE C) (QUOTE #III))
;
;	'((HANOI LAMBDA (A B C N) (COND ((NULL N) NIL)
;		(*T* (APPEND3 (HANOI A C B (CDR N))
;			(CONS (CONS A (CONS B NIL)) NIL)
;				(HANOI C B A (CDR N))))))
;
;	(APPEND2 LAMBDA (A B)
;		(COND ((NULL A) B)
;			(*T* (CONS (CAR A) (APPEND2 (CDR A) B)))))
;
;	(APPEND3 LAMBDA (A B C)
;		(APPEND2 A (APPEND2 B C)))))

:L LIB/ASSOC.L
:L LIB/PAIR.L
:L LIB/CAAR.L
;L LIB/CADR.L
;L LIB/CDDR.L
:L LIB/CAAAR.L
;L LIB/CAADR.L
;L LIB/CADAR.L
;L LIB/CADDR.L
:L LIB/CAAAAR.L
;L LIB/CADADR.L
;L LIB/CADDAR.L
:L LIB/APPEND.L
:L LIB/MEMBER.L

; ---code---
(DEFUN XEVAL (X E)
	(LABEL

	; Evaluate each member of list X in environment E:
	; (M1 ... Mn) -> (eval[M1] ... eval[Mn])
	((EVLIST (LAMBDA (X E)
		(COND ((NULL X) NIL)
			(T (CONS (_EVAL (CAR X) E) (EVLIST (CDR X) E))))))

	; Check whether the atom X is a keyword.
	(SYNTAXP (LAMBDA (X)
		(MEMBER X '(BOTTOM QUOTE CAR CDR CONS ATOM NULL EQ COND
					LIST IMPLODE EXPLODE LAMBDA LABEL))))

	; Find the value of a symbol
	(VALUE (LAMBDA (X E)
		(LABEL ((B (ASSOC X E)))
			(COND ((NULL B) NIL)
				(T (CDR B))))))

	; Create a local environment from a LABEL
	(MAKEENV (LAMBDA (LBL ENV)
		(COND ((NULL LBL) ENV)
			(T (MAKEENV (CDR LBL)
				(CONS (CONS (CAAR LBL) (_EVAL (CADAR LBL) ENV)) ENV))))))

	(_EVAL (LAMBDA (X E)
		; --- Atoms ---
		(COND ((ATOM X)
				(COND ((NULL X) NIL)	; empty list
					((EQ X *T*) *T*)	; predefined symbols
					((EQ X *F*) *F*)
					((EQ X 'NIL) NIL)
					((EQ X 'T) *T*)
					(T (VALUE X E))))	; default: retrieve binding from E
			; --- Lists with atom at first slot ---
			((ATOM (CAR X))
				; first try the literal keywords
				(COND ((EQ (CAR X) 'BOTTOM)
						(BOTTOM (CADR X)))
					((EQ (CAR X) 'QUOTE)
						(CADR X))
					((EQ (CAR X) 'CAR)
						(CAR (_EVAL (CADR X) E)))
					((EQ (CAR X) 'CDR)
						(CDR (_EVAL (CADR X) E)))
					((EQ (CAR X) 'CONS)
						(CONS (_EVAL (CADR X) E) (_EVAL (CADDR X) E)))
					((EQ (CAR X) 'ATOM)
						(ATOM (_EVAL (CADR X) E)))
					((EQ (CAR X) 'NULL)
						(NULL (_EVAL (CADR X) E)))
					((EQ (CAR X) 'EQ)
						(EQ (_EVAL (CADR X) E) (_EVAL (CADDR X) E)))
					((EQ (CAR X) 'COND)
						(COND ((EQ (_EVAL (CAADR X) E) *T*)
								(_EVAL (CADADR X) E))
							((EQ (_EVAL (CAADR X) E) *F*)
								(_EVAL (CONS 'COND (CDDR X)) E))
							(T (BOTTOM 'XEVAL:COND))))
					((EQ (CAR X) 'LIST)
						(EVLIST (CDR X) E))
					((EQ (CAR X) 'IMPLODE)
						(IMPLODE (_EVAL (CADR X) E)))
					((EQ (CAR X) 'EXPLODE)
						(EXPLODE (_EVAL (CADR X) E)))
					; lambda functions already are in their normal forms
					((EQ (CAR X) 'LAMBDA)
						X)
					((EQ (CAR X) 'LABEL)
						(_EVAL (CADDR X) (MAKEENV (CADR X) E)))
					; At this point, the first member must be a variable.
					; If it is bound to a keyword, evaluate it and try
					; again.
					((SYNTAXP (_EVAL (CAR X) E))
						(_EVAL (CONS (_EVAL (CAR X) E) (CDR X)) E))
					; If it is bound to a non-keyword atom, this is an
					; application of a non-function
					((ATOM (_EVAL (CAR X) E))
						(BOTTOM 'XEVAL:NON-FUNCTION))
					; By default, replace the variable with its binding
					; and try again
					(T
						(_EVAL (CONS (_EVAL (CAR X) E) (CDR X)) E))))
			; --- Applications of lambda functions ---
			((EQ (CAAR X) 'LAMBDA)
				; Bind the actual arguments to the formal arguments
				; and then evaluate the term of the lambda function
				; in the union of the new bindings and E
				(_EVAL (CADDAR X)
					(APPEND (PAIR (CADAR X) (EVLIST (CDR X) E)) E)))
			; --- Default: undefined ---
			(T (BOTTOM 'XEVAL))))))

	(_EVAL X E)))
